"""
Parser for reading some graph format that I'm currently unable to read with Sage
or NetworkX.
"""
import networkx
import string
from sage.sets.set import Set_generic

class GraphIO:

    def __init__(self):
        """
        """

    def __dealloc__(self):
        """
        """

    def improve_labels(self, G, relabel=False):
        """
        """
        if relabel:
            G.relabel(inplace=True)

        elif all(isinstance(u, unicode) for u in G):
            myaction = string.atoi
        elif all(isinstance(u, str) for u in G):
            myaction = ZZ

        try:
            L = dict( (u,myaction(u)) for u in G )
        except:
            L = dict( (u,str(u)) for u in G )
        G.relabel(perm=L, inplace=True)

        return G


    def read_graph(self, path, filename, format=None, directed=False, relabel=False, verbose=False):
        """
        Read the (di)graph from filename.

        If the file format is not specified, guess the format and try different
        methods. The result could be weird.
        """
        dfunk = {
            'graphml': self.read_graphml,
            'edgelist': self.read_edgelist,
            'net': self.read_net_file,
            'dimacs': self.read_dimacs
            }

        if format:
            if format in dfunk:
                D = dfunk[format](filename, directed=directed, relabel=relabel)
            else:
                raise ValueError("Unknown format: "+str(format))

        else:
            # Guess format from filename
            extension = filename.split()[-1]
            if extension in dfunk:
                D = dfunk[extension](path+'/'+filename, directed=directed, relabel=relabel)
            else:
                # Try all known format
                for extension in dfunk:
                    try:
                        D = dfunk[extension](path+'/'+filename, directed=directed, relabel=relabel)
                        if D:
                            if verbose:
                                print("Suspected format:",extension)
                            break
                    except:
                        continue
                else:
                    raise ValueError("Unknwon file format or unexpected error with file "+str(filename))
        D.name(filename)
        return D

    def read_graphml(self, filename, directed=False, relabel=False):
        """
        Read graphml file format as used in http://www.graphdrawing.org/data/
        for the Rome dataset

        The input parameter 'directed' is ignored
        """
        try:
            f = open(filename, 'r')
        except:
            raise ValueError('Unable to open file '+filename)

        # We search for graph type: directed or undirected
        for line in f:
            if line.startswith('<graph') and not line.startswith('<graphml'):
                if 'undirected' in line:
                    G = Graph()
                else: # if 'directed' in line: or not specified
                    G = DiGraph()
                break

        # We set graph name
        G.name(filename.split('/')[-1].split('.graphml')[0])

        # We read nodes and edges
        for line in f:
            if line.startswith('<node'):
                a,_,l = line.partition('"')
                id,_,l = l.partition('"')
                if id and line==a+'"'+id+'"'+l:
                    G.add_vertex(id)
                else:
                    raise ValueError('Unrecognized entry: '+line)
            elif line.startswith('<edge'):
                _,_,l = line.partition('source="')
                u,_,l = l.partition('" target="')
                v,_,_ = l.partition('"')
                if u in G and v in G:
                    G.add_edge(u,v)
                else:
                    print('u =',u,'v =',v)
                    raise ValueError('Someting goes wrong while reading line: '+line)

        f.close()

        return self.improve_labels(G, relabel=relabel)


    def read_edgelist(self, filename, directed=False, relabel=False):
        """
        Read edgelist file format
        """
        import networkx
        try:
            for delimiter in [' ', '\t', '|']:
                try:
                    if directed:
                        G = DiGraph( networkx.read_edgelist(filename, delimiter=delimiter,
                                                            create_using=networkx.DiGraph()) )
                    else:
                        G = Graph( networkx.read_edgelist(filename, delimiter=delimiter) )
                    if G: 
                        break
                except:
                    continue

        except:
            print('Unable to open file '+filename)
            G = Graph()

        return self.improve_labels(G, relabel=relabel)


    def read_net_file(self, input, directed=True, relabel=False):
        """
        Read and return the digraph from a .net file as used in the "Bayesian
        Network Repository".

        http://www.bnlearn.com/bnrepository/

        The input parameter 'directed' is ignored
        """
        try:
            with open(input, 'r') as f:
                D = DiGraph()
                for line in f:
                    # arcs are in lines starting with 'potential'
                    if not line.startswith('potential'):
                        continue
                    _,_,s = line.partition('(')
                    s,_,_ = s.partition(')')
                    s = s.replace('| ','').strip()
                    L = s.split()
                    if len(L)<=1:
                        continue
                    # Adjacency is given as "destination | source1, source2, etc.'.
                    v = L[0]
                    for u in L[1:]:
                        D.add_edge(u,v)
                f.close()

            D.name(input[:-4])

        except:
            print('Unable to open file '+input)
            D = DiGraph()

        return self.improve_labels(D, relabel=relabel)


    def read_dimacs(self, filename, directed=True, relabel=False):
        """
        Read DIMACS file format:

        c FILE: jean.col
        c Translated from Stanford GraphBase File: jean.gb
        c Stanford GraphBase ID: book(?jean?,80,0,1,356,0,0,0)
        p edge 80 508
        e 1 14
        e 2 37
        e 2 75

        The input parameter 'directed' is ignored
        """
        try:
            with open(filename, 'r') as f:
                D = DiGraph()
                D.name(filename)
                for line in f:
                    if line.startswith('c FILE:'):
                        _,_,gname = line.partition('FILE:')
                        D.name(gname.strip())
                    elif line.startswith('e '):
                        u,v = (line[2:]).split()
                        D.add_edge(u,v)
                    else:
                        continue
                f.close()

        except:
            print('Unable to open file '+input)
            D = DiGraph()

        return self.improve_labels(D, relabel=relabel)


    def read_hg(self, filename, directed=False, relabel=False):
        """
        Read hyperbolic graph (hg) file format. 

        This format is used by Krioukov et al.
        http://named-data.github.io/Hyperbolic-Graph-Generator/

        The first line contains all the parameters used to generate the graph.
        The next N lines gives the coordinates of the vertices.
        The next lines are the edges in edgelist format.

        N       1000    T       0.00000 G       2.00000 K       10.00000        Z       1.00000 S       1       I       1
        1       14.11310        6.26550
        2       15.20478        5.85943
        3       1.11295 0.80503
        4       13.47119        6.27716
        ...
        1000    15.34961        0.20229
        1       3
        1       251
        1       539
        1       745
        2       3
        2       251
        2       562
        ...
        965     966
        965     977
        966     977

        The input parameter 'directed' is ignored
        """
        def _parse_parameters(st):
            """
            """
            L = st.split()
            res = dict( ( L[i], (ZZ(L[i+1]) if L[i]=='N' else RR(L[i+1])) ) for i in range(0, len(L), 2) )
            return res

        try:
            with open(filename, 'r') as f:
                G = Graph()
                # Read the first line containing all parameters
                param = _parse_parameters( f.readline() )

                # Read the N next lines containing coordinates
                pos = dict()
                for i in range(param['N']):
                    u,x,y = f.readline().split()
                    pos[u] = (RR(x), RR(y))
                G.add_vertices( u for u in pos )
                G.set_pos( pos )

                # Read the edges
                for line in f:
                    u,v = line.split()
                    G.add_edge(u, v)

                # set the name of the graph
                G.name(filename)

                f.close()

        except:
            print('Unable to open file '+input)
            G = Graph()

        return self.improve_labels(G, relabel=relabel)



    def write_edgelist(self, G, filename, comments=''):
        """
        Write input graph as an edgelist
        """
        try:
            f = open(filename, 'w')
        except:
            raise ValueError('Unable to open file: '+filename)

        f.write('# Graph name: '+G.name()+'\n')
        f.write('# '+('directed' if G.is_directed() else 'undirected')+'\n')
        f.write('# N = '+str(G.order())+'\n# M = '+str(G.size())+'\n')
        if comments:
            f.write('# '+comments+'\n')
        my_map = {}
        sep = ' '
        for u in G:
            if isinstance(u, str):
                if ' ' in u or "'" in u or '"' in u:
                    sep = '\t' if not '\t' in u else '|'
                if u.isdigit():
                    my_map[u] = '"'+str(u)+'"'
                else:
                    my_map[u] = str(u)
            elif isinstance(u, (tuple, list, set, frozenset, Set_generic)):
                sep = '\t'
                my_map[u] = str(u)
            else:
                my_map[u] = str(u)
                
        for u,v in G.edge_iterator(labels=None):
            f.write(my_map[u]+sep+my_map[v]+'\n')
        f.write('# end\n')
        f.close()


    def convert_graphml_to_edgelist(self, sourcefile, targetfile, relabel=True):
        """
        """
        self.write_edgelist(self.read_graphml(sourcefile, relabel=relabel), targetfile)


    def convert_graphml_to_edgelist_dir(self, sourcedir='', targetdir='', relabel=True):
        """
        """
        listing = os.listdir(sourcedir) # get all files in sourcedir
        for fname in listing:
            if fname.endswith('.graphml'):
                G = self.read_graphml(sourcedir+'/'+fname, relabel=relabel)
                self.write_edgelist(G, targetdir+'/'+G.name()+'.edgelist')


    def convert_files(self, sourcedir='', targetdir='', ext1='', ext2='', relabel=False):
        if ext1=='graphml':
            reader = self.read_graphml
        elif ext1=='net':
            reader = self.read_net_file
        elif ext1.lower() in ['dimacs', 'col']:
            reader = self.read_dimacs
        else:
            reader = self.read_edgelist

        if ext2=='edgelist':
            writer = self.write_edgelist
        else:
            raise ValueError('Unknown format: '+ext2)

        for fname in os.listdir(sourcedir):
            if fname.endswith(ext1):
                dname = fname.replace('.'+ext1, '.'+ext2)
                print('convert '+fname+' to '+dname)
                G = reader(sourcedir+'/'+fname, relabel=relabel)
                writer(G, targetdir+'/'+dname)
        


