"""
Ordering formulations:

All different, ``A``, where variable `x_u^i=1` if vertex `u` is at position `i` and `x_u^i=0` otherwise::

    .. MATH:
        :nowrap:

        \begin{alignat*}{2}
        \sum_{i=0}^{n-1} x_u^i & = 1 &\qquad \forall u\in V\\
        \sum_{u\in V} x_u^i & = 1 & \qquad \forall i\in[0,n-1]
        \end{alignat*}


All different with precedence, ``P``, where variable `x_u^i=1` if vertex `u` is at a position `j\in[0,i]` and `x_u^i=0` otherwise::

    .. MATH:
        :nowrap:

        \begin{alignat*}{2}
        x_u^i & \leq x_u^{i+1} & \qquad \forall u\in V,\ \forall i\in[0,n-2]\\
        \sum_{v\in V} x_u^i & = i+1 &\qquad \forall i\in[0,n-1]
        \end{alignat*}

All different with precedence, ``Q``, where variable `x_u^i=1` if vertex `u` is at a position `j\in[0,i]` and `x_u^i=0` otherwise::

    .. MATH:
        :nowrap:

        \begin{alignat*}{2}
        \sum_{j=0}^{i-1}x_u^j & \leq i\cdot x_u^{i} & \qquad \forall u\in V,\ \forall i\in[0,n-1]\\
        \sum_{v\in V} x_u^i & = i+1 &\qquad \forall i\in[0,n-1]
        \end{alignat*}


        
Strict ordering, ``S``, where variable ``x_{u,v}=1`` if ``u`` is before ``v`` in the ordering and 0 otherwise::

    .. MATH:
        :nowrap:

        \begin{alignat*}{2}
        x_{u,v} + x_{v,u} & =1 & \qquad \forall u,v\in V\\
        x_{u,v} + x_{v,w} - x_{u,w} & \leq 1 &\qquad \forall u,v,w in V
        \end{alignat*}



REFERENCES:

.. [LL+14] M. C. Lopez-Loces, N. Castillo-Garcia, H. J. Fraire H., P. Bouvry,
  J. E. Pecero, R. A. Pazos Rangel, J. J. G. Barbosa, and F. Valdez. *A new
  integer linear programming model for the cutwidth minimization problem of a
  connected undirected graph*. In O. Castillo, P. Melin, W. Pedrycz, and
  J. Kacprzyk, editors, Recent Advances on Hybrid Approaches for Designing
  Intelligent Systems, volume 547 of Studies in Computational Intelligence,
  pages 509–517. Springer, 2014.

.. [LMSY05] J. Luttamaguzi, M. J. Pelsmajer, Z. Shen, and B. Yang. *Integer
  programming methods for several optimization problems in graph theory*. In
  G. Hu, editor, 20th International Conference on Computers and Their
  Applications (CATA), pages 50–55. ISCA, Mar. 2005.
"""
from sage.structure.sage_object import SageObject
from sage.numerical.mip import MixedIntegerLinearProgram

class GraphLayout:
    """
    """

    def __init__(self, G, solver=None, timelimit=None, verbose=0, check=False, logfile=None, nthreads=None):
        """
        """
        from sage.graphs.graph import Graph
        if not isinstance(G, Graph):
            raise ValueError("The first input parameter must be a Graph.")

        self._N = G.order()
        self.G  = G.copy()
        self._V = G.vertices()
        self._E = G.edges(labels=None)
        self._solver    = solver
        self._timelimit = timelimit
        self._verbose   = verbose
        self._check = check
        self.logfile = logfile
        self._nthreads = nthreads
        self.mip = None
        self.time = +oo
        self._version = None

    def __dealloc__(self):
        """
        """
        del self.mip

    def _repr_(self):
        r"""
        Return ``self`` as a unique str.
        """
        return "Tool for computing linear orderings of graph '{}'".format(self.G)


    def _init_mip(self):
        """
        Initialize the MIP and the opt variable, and set solver parameters
        """
        if not self.mip is None:
            del self.mip

        self.mip = MixedIntegerLinearProgram( maximization = False, solver = self._solver )
        self.opt = self.mip.new_variable(integer=True, nonnegative=True, name="opt")


        if not self._timelimit is None:
            self.mip.solver_parameter("timelimit", self._timelimit)

        if not self.logfile is None:
            self.mip.solver_parameter("logfile", self.logfile)

        if not self._nthreads is None:
            self.mip.solver_parameter("CPX_PARAM_THREADS", self._nthreads)
            
        self.objective_value = -1


    def _set_bounds(self, lowerbound=None, upperbound=None):
        """
        """
        if not lowerbound is None:
            self.mip.add_constraint( self.opt['opt'] >= lowerbound )
        if not upperbound is None:
            self.mip.add_constraint( self.opt['opt'] <= upperbound )
        
    def _set_objective(self):
        """
        """
        self.mip.set_objective(self.opt['opt'])


    def _solve(self):
        """
        """
        try:
            t = walltime()
            obj = self.mip.solve( log=self._verbose )
            self.time = walltime()-t
            self.objective_value = int(round( self.mip.get_values(self.opt)['opt'] ))
        except:
            # Typical problem: the primal has no feasible solution
            print("WARNING: nofeasible solution for {} with {}".format(self.G.name(),self._version))
            self.time = self._timelimit
            self.objective_value = +oo
        
    def _all_different(self, version):
        """
        Initialize the x binary variables and add the requested all diff constraints
        """
        self._version = version
        self.x = self.mip.new_variable(binary=True, nonnegative=True, name="x")
        
        if version[0]=='A':
            for v in self._V:
                self.mip.add_constraint( self.mip.sum( self.x[v,i] for i in range(self._N) ) == 1 )
            for i in range(self._N):
                self.mip.add_constraint( self.mip.sum( self.x[v,i] for v in self._V ) == 1 )

        elif version[0]=='P':
            for v in self._V:
                for i in range(self._N-1):
                    self.mip.add_constraint( self.x[v,i] <= self.x[v,i+1] )
                self.mip.add_constraint( self.x[v,self._N-1] == 1 )
            for i in range(self._N):
                self.mip.add_constraint( self.mip.sum( self.x[v,i] for v in self._V ) == i+1)

        elif version[0]=='Q':
            for v in self._V:
                for i in range(self._N-1):
                    self.mip.add_constraint( self.mip.sum( self.x[v,k] for k in range(i) ) <= i*self.x[v,i] )
                self.mip.add_constraint( self.x[v,self._N-1] == 1 )
            for i in range(self._N):
                self.mip.add_constraint( self.mip.sum( self.x[v,i] for v in self._V ) == i+1 )

        elif version[0]=='R':
            for v in self._V:
                for i in range(self._N-1):
                    for j in range(i+1,self._N):
                        self.mip.add_constraint( self.x[v,i] <= self.x[v,j] )
                self.mip.add_constraint( self.x[v,self._N-1] == 1 )
            for i in range(self._N):
                self.mip.add_constraint( self.mip.sum( self.x[v,i] for v in self._V ) == i+1 )

        elif version[0]=='S':
            import itertools
            for u,v in itertools.combinations(self._V, 2):
                self.mip.add_constraint( self.x[u,v]+self.x[v,u] == 1 )
            for u,v,w in itertools.permutations(self._V, 3):
                self.mip.add_constraint( self.x[u,v]+self.x[v,w]-self.x[u,w] <= 1 )
                
        else:
            raise ValueError("Unknown version for the all_different method.")


    def _define_lambdas(self, version):
        """
        """
        if version[0] in ['A']:
            self._lambda_tau = lambda _u,_v,_i:self.mip.sum( self.x[_v,_j]-self.x[_u,_j] for _j in range(_i+1) )
            self._lambda_ell = lambda _u,_v:self.mip.sum( _j*(self.x[_u,_j]-self.x[_v,_j]) for _j in range(self._N) )
            self._lambda_pos = lambda _u:self.mip.sum( _j*self.x[_u,_j] for _j in range(self._N) )

        elif version[0] in ['P','Q','R']:
            self._lambda_tau = lambda _u,_v,_i:self.x[_v,_i]-self.x[_u,_i]
            self._lambda_ell = lambda _u,_v:self.mip.sum( self.x[_v,_j]-self.x[_u,_j] for _j in range(self._N) )
            self._lambda_pos = lambda _u:self._N-self.mip.sum( self.x[_u,_j] for _j in range(self._N) )

        elif version[0] in ['S']:
            self._lambda_tau = lambda _u_,_v_,_i_:0
            self._lambda_ell = lambda _u,_v:self.mip.sum( self.x[_w,_u] for _w in self._V if _w!=_u ) - self.mip.sum( self.x[_w,_v] for _w in self._V if _w!=_v )
            self._lambda_pos = lambda _u:self.mip.sum( self.x[_w,_u] for _w in self._V if _w!=_u )

        else:
            raise ValueError("Unknown version {}".format(version))


    def extract_ordering(self, version):
        """
        """
        try:
            val_x = self.mip.get_values( self.x )
        except:
            # This happens when no feasible solution is found
            return self.G.vertices()

        seq = [0]*self._N

        if version[0] in ['A']:
            for u in self._V:
                seq[int(round(sum(i*val_x[u,i] for i in range(self._N))))] = u
            if sum(val_x[u,i] for u in self._V for i in range(self._N))>self._N:
                raise ValueError("Something is freaky wrong here!")

            if version[1:] in ['Duarte12', 'Gurski15']:
                seq = seq[::-1]
                
        elif version[0] in ['P', 'Q', 'R']:
            for u in self._V:
                seq[self._N - int(round(sum(val_x[u,i] for i in range(self._N))))] = u

        elif version[0] in ['S']:
            cpt = {u:int(sum(val_x[w,u] for w in self._V if w!=u)) for u in self._V}
            for u,pos in cpt.items():
                seq[pos] = u

        else:
            raise ValueError("Unknown version: {}".format(version))


        return seq


    def get_relative_mip_gap(self, digits=4):
        """
        Return the optimality gap of the best known solution.
        """
        if digits<0:
            raise ValueError("Parameter digits must a nonnegative integer.")
        if self._version == 'bwBBNNPR':
            return round(self._relative_mip_gap, digits)
        
        try:
            gap = round(self.mip.get_relative_objective_gap(), digits)
        except:
            # the method is not yet implemented or something goes wrong
            gap = +oo
        return gap

    def get_best_objective_value(self, digits=4):
        """
        Return the value of the currently best known bound.
        """
        if digits<0:
            raise ValueError("Parameter digits must a nonnegative integer.")
        if self._version == 'bwBBNNPR':
            return round(self._best_objective_value, digits)
        
        try:
            bound = round(self.mip.best_known_objective_bound(), digits)
        except:
            # the method is not yet implemented or something goes wrong
            bound = 0   # here all optimization problems are minimization stuff
        return bound



    
    def pathwidth(self, version, lowerbound=None, upperbound=None):
        """
        """
        self._init_mip()
        self._all_different(version)
        self._define_lambdas(version)

        y = self.mip.new_variable( name        = "y",
                                    binary      = version[1:] in ['D3', 'SoPi10', 'Gurski15', 'Duarte12'],
                                    nonnegative = True)


        if (version[1:]=='D1' and version[0] in ['A','P','Q','R']):
            for i in range(self._N-1):
                for u in self._V:
                    for v in self.G.neighbor_iterator(u):
                        self.mip.add_constraint( self._lambda_tau(u,v,i) <= y[u,i] )

            for i in range(self._N-1):
                self.mip.add_constraint( self.mip.sum( y[u,i] for u in self._V ) <= self.opt['opt'] )


        elif version[0] in ['P', 'Q', 'R'] and version[1:] in ['SoPi10']:
            varu = self.mip.new_variable( name="u", binary=True, nonnegative=True )
            for i in range(self._N-1):
                for u in self._V:
                    self.mip.add_constraint( self.mip.sum( y[v,i] for v in self.G.neighbors(u) )
                                                >= self.G.degree(u)*self.x[u,i] )
                    self.mip.add_constraint( self.x[u,i] <= y[u,i] )
                    self.mip.add_constraint( y[u,i] - self.x[u,i] <= varu[u,i] )
            
            for i in range(self._N-1):
                self.mip.add_constraint( self.mip.sum( varu[u,i] for u in self._V ) <= self.opt['opt'] )

                
        elif version[0] in ['P', 'Q', 'R'] and version[1:] in ['D2', 'D3']:
            for i in range(self._N-1):
                for u in self._V:

                    if version[1:] in ['D2']:
                        for v in self.G.neighbor_iterator(u):
                            self.mip.add_constraint( self.x[v,i] <= y[u,i] )
                        self.mip.add_constraint( self.x[u,i] <= y[u,i] )

                    else: # version[1:] in ['D3']
                        self.mip.add_constraint( self.mip.sum( self.x[v,i] for v in self.G.neighbors(u) )
                                                <= self._N*y[u,i] )
                        self.mip.add_constraint( self.x[u,i] <= y[u,i] )

            for i in range(self._N-1):
                self.mip.add_constraint( self.mip.sum( y[u,i]-self.x[u,i] for u in self._V )
                                            <= self.opt['opt'] )


        elif version[0] in ['S']:
            ll = self.mip.new_variable( name = "l", binary = True, nonnegative = True)
            rr = self.mip.new_variable( name = "l", binary = True, nonnegative = True)
            for i in range(self._N):
                for u in self._V:
                    self.mip.add_constraint( i+1 - self.mip.sum( self.x[w,u] for w in self._V if u!=w )
                                             <= self._N*ll[u,i] )
                    self.mip.add_constraint( self.mip.sum( self.x[w,u] for w in self._V if u!=w ) -i
                                             <= self._N*rr[u,i] )
                    for v in self.G.neighbors(u):
                        self.mip.add_constraint( ll[v,i]+rr[u,i]-1 <= y[u,i] ) 

                self.mip.add_constraint( self.mip.sum( y[u,i] for u in self._V ) <= self.opt['opt'] )


        elif version in ['AGurski15', 'ADuarte12']:
            import itertools
            YY = self.mip.new_variable(name="Y", binary=True, nonnegative=True)
            for u,v in self._E:
                for k,l in itertools.permutations(range(self._N),2):
                    self.mip.add_constraint( y[u,k,v,l] <= self.x[u,k] )
                    self.mip.add_constraint( y[u,k,v,l] <= self.x[v,l] )
                    self.mip.add_constraint( self.x[u,k]+self.x[v,l] <= y[u,k,v,l] + 1 )

            for u,v in self._E:
                for k,l in itertools.combinations(range(self._N),2):
                    for i in range(k,l):
                        self.mip.add_constraint( y[u,k,v,l] <= YY[k,i] )
                        self.mip.add_constraint( y[u,l,v,k] <= YY[k,i] )

            for i in range(self._N-1):
                self.mip.add_constraint( self.mip.sum( YY[k,i] for k in range(i+1) ) <= self.opt['opt'] )

                
        else:
            raise ValueError("Unknown version for pathwidth: {}.".format(version))

                    
        self._set_bounds(lowerbound=lowerbound, upperbound=upperbound)
        self._set_objective()
        self._solve()

        if self._check and not self._check_pathwidth(version):
            raise ValueError("pathwidth version {}: something goes wrong...".format(version))
        return self.objective_value, self.extract_ordering(version)

    def _check_pathwidth(self, version):
        """
        """
        pw  = self.objective_value
        L   = self.extract_ordering(version)

        pwL = 0
        S = set()
        neighbors_of_S_in_V_minus_S = set()

        for u in L:
            # We remove u from the neighbors of S
            neighbors_of_S_in_V_minus_S.discard(u)
    
            # We add vertex u to the set S
            S.add(u)

            # We add the (out-)neighbors of u to the neighbors of S
            for v in self.G.neighbor_iterator(u):
                if not v in S:
                    neighbors_of_S_in_V_minus_S.add(v)

            # We update the cost of the vertex separation
            pwL = max( pwL, len(neighbors_of_S_in_V_minus_S) )

        # These two values should be equal, but is some cases, the objective
        # value returned by the solver (best known integer solution) is higher
        # than the cost of the current ordering. This can be observed with graph
        # ``~/Recherche/Instances/CMPLIB_2010/small/p100_24_34.edgelist``, model
        # ``'SS'``, and ``timelimit=10`` where reported objective value is 24
        # while the current solution has cost 14. O_o
        return pwL<=pw

    

    def cutwidth(self, version, lowerbound=None, upperbound=None):
        """
        """
        self._init_mip()
        self._all_different(version)
        self._define_lambdas(version)

        y = self.mip.new_variable( name = "y", nonnegative = True,
                                    integer = version in ['SD1'] or version[1:]=='CK',
                                    binary = version in ['ALPSY05', 'SLL+14'] )

        
        if version[1:] in ['D1'] and version[0] in ['A','P','Q','R']:
            for u,v in self._E:
                for i in range(self._N):
                    self.mip.add_constraint( self._lambda_tau(u,v,i) <= y[u,v,i] )
                    self.mip.add_constraint( self._lambda_tau(v,u,i) <= y[u,v,i] )

            for i in range(self._N):
                self.mip.add_constraint( self.mip.sum( y[u,v,i] for u,v in self._E ) <= self.opt['opt'] )


        elif version[1:] in ['D1'] and version[0] in ['S']:
            ll = self.mip.new_variable( name = "l", binary = True, nonnegative = True)
            rr = self.mip.new_variable( name = "l", binary = True, nonnegative = True)
            for i in range(self._N):
                for u in self._V:
                    self.mip.add_constraint( i+1 - self.mip.sum( self.x[w,u] for w in self._V if u!=w )
                                                <= self._N*ll[u,i] )
                    self.mip.add_constraint( self.mip.sum( self.x[w,u] for w in self._V if u!=w ) -i
                                                <= self._N*rr[u,i] )
                        
            for i in range(self._N):
                for u,v in self._E:
                    self.mip.add_constraint( ll[u,i]+rr[v,i]-1 <= y[u,v,i] ) 
                    self.mip.add_constraint( ll[v,i]+rr[u,i]-1 <= y[u,v,i] ) 
            
            for i in range(self._N):
                self.mip.add_constraint( self.mip.sum( y[u,v,i] for u,v in self._E ) <= self.opt['opt'] )


                
        elif version[1:] in ['CK'] and version[0] in ['A','P','Q','R']: # version MILP-20140424
            if version[0]=='A':
                is_at_pos = lambda _v,_i:self.x[_v,_i]
                is_before = lambda _u,_i:self.mip.sum( self.x[_u,_j] for _j in range(_i) )
            else:
                is_at_pos = lambda _v,_i:(self.x[_v,_i]-self.x[_v,_i-1])
                is_before = lambda _u,_i:self.x[_u,_i-1]
            
            diff  = self.mip.new_variable(integer=True, nonnegative=False, name='diff')
            deg   = self.G.degree(labels=True)
            delta = max(deg.values())
            self.mip.add_constraint( self.mip.sum( deg[v]*self.x[v,0] for v in self._V ) == y[0] )
            for i in range(1, self._N):
                for v in self._V:
                    self.mip.add_constraint( deg[v]*is_at_pos(v,i)
                                            - 2*self.mip.sum( is_before(u,i) for u in self.G.neighbors(v) )
                                            - 2*delta*(1-is_at_pos(v,i))
                                            <= diff[i] )

                self.mip.add_constraint( y[i-1] + diff[i] == y[i] )

            self.mip.add_constraint( y[self._N-1] == 0 )

            for i in range(self._N):
                self.mip.add_constraint( y[i] <= self.opt['opt'] )



        elif version in ['ALPSY05', 'AMPDP13']:
            import itertools
            for u,v in self._E:
                for k,l in itertools.permutations(range(self._N),2):
                    self.mip.add_constraint( y[u,k,v,l] <= self.x[u,k] )
                    self.mip.add_constraint( y[u,k,v,l] <= self.x[v,l] )
                    self.mip.add_constraint( self.x[u,k] + self.x[v,l] - y[u,k,v,l] <= 1 )

            for i in range(self._N-1):
                self.mip.add_constraint( self.mip.sum( self.mip.sum( y[u,k,v,l]+y[u,l,v,k] for k in range(i+1) for l in range(i+1,self._N) ) for u,v in self._E ) <= self.opt['opt'] )



        elif version=='SLL+14': # improved version with nm variables instead of n^3
            y = self.mip.new_variable(binary=True, nonnegative=True)
            p = self.mip.new_variable(integer=True, nonnegative=True)
            R = self.mip.new_variable(binary=True, nonnegative=True)
            delta = self.mip.new_variable(binary=True, nonnegative=True)

            # Section 2.3.2 of [LL+14]_
            for u,v in self._E:
                self.mip.add_constraint( y[u,v] == self.x[u,v] )
                self.mip.add_constraint( y[v,u] == self.x[v,u] )

            # Section 2.3.3 of [LL+14]_: indicates position of u
            for u in self._V:
                self.mip.add_constraint( self._N - self.mip.sum( self.x[u,v] for v in self._V if u!=v ) == p[u] )

            # Section 2.3.4 of [LL+14]_:
            for u in self._V:
                for i in range(1,self._N):
                    self.mip.add_constraint( i+1 - self._N*(1-R[u,i]) <= p[u] )
                    self.mip.add_constraint( p[u]+1 - (i+1) <= (self._N-1)*R[u,i] )

            # Section 2.3.5 of [LL+14]_: Modified version
            for u,v in self._E:
                for i in range(1,self._N):
                    self.mip.add_constraint( y[u,v] + (1-R[u,i]) + R[v,i] >= 3*delta[u,v,i] )
                    self.mip.add_constraint( y[u,v] + (1-R[u,i]) + R[v,i] <= delta[u,v,i] + 2 )

                    self.mip.add_constraint( y[v,u] + (1-R[v,i]) + R[u,i] >= 3*delta[v,u,i] )
                    self.mip.add_constraint( y[v,u] + (1-R[v,i]) + R[u,i] <= delta[v,u,i] + 2 )

            for i in range(1,self._N):
                self.mip.add_constraint( self.mip.sum( delta[u,v,i]+delta[v,u,i] for u,v in self._E )
                                         <= self.opt['opt'] )


        else:
            raise ValueError("Unknown version: {}.".format(version))

        self._set_bounds(lowerbound=lowerbound, upperbound=upperbound)
        self._set_objective()
        self._solve()

        if self._check and not self._check_cutwidth(version):
            raise ValueError("cutwidth version {}: something goes wrong...".format(version))
        return self.objective_value, self.extract_ordering(version)

    def _check_cutwidth(self, version):
        """
        """
        cw  = self.objective_value
        L   = self.extract_ordering(version)
        pos = {u:i for i,u in enumerate(L)}
        cpt = [0]*len(L)
        for u,v in self._E:
            x,y = pos[u],pos[v]
            if x>y:
                x,y = y,x
            for i in range(x,y):
                cpt[i] += 1

        return max(cpt)<=cw


    def bandwidth(self, version, lowerbound=None, upperbound=None):
        """
        """
        self._init_mip()
        self._all_different(version[0])
        self._define_lambdas(version)

    
        if version[1:] in ['D1'] and version[0] in ['A','P','Q','R','S']:
            for u,v in self._E:
                self.mip.add_constraint( self._lambda_ell(u,v) <= self.opt['opt'] )
                self.mip.add_constraint( self._lambda_ell(v,u) <= self.opt['opt'] )


        elif version[1:] in ['Pinana04'] and version[0] in ['A','P','Q','R','S']:
            pos = self.mip.new_variable(name="pos", integer=True, nonnegative=True)
            for u in self._V:
                self.mip.add_constraint( self._lambda_pos(u) == pos[u] )
            for u,v in self._E:
                self.mip.add_constraint( pos[u]-pos[v] <= self.opt['opt'] )
                self.mip.add_constraint( pos[v]-pos[u] <= self.opt['opt'] )


        elif version=='ALPSY05':
            import itertools
            y = self.mip.new_variable(name="y", binary=True, nonnegative=True)
            for u,v in self._E:
                for k,l in itertools.permutations(range(self._N),2):
                    self.mip.add_constraint( y[u,k,v,l] <= self.x[u,k] )
                    self.mip.add_constraint( y[u,k,v,l] <= self.x[v,l] )
                    self.mip.add_constraint( self.x[u,k]+self.x[v,l] <= y[u,k,v,l] + 1 )

            for u,v in self._E:
                for k,l in itertools.permutations(range(self._N),2):
                    self.mip.add_constraint( abs(k-l)*y[u,k,v,l] <= self.opt['opt'] )

                    
        else:
            raise ValueError("Unknown version: {}.".format(version))


        self._set_bounds(lowerbound=lowerbound, upperbound=upperbound)
        self._set_objective()
        self._solve()

        if self._check and not self._check_bandwidth(version):
            raise ValueError("bandwidth version {}: something goes wrong...".format(version))
        return self.objective_value, self.extract_ordering(version)

    def _check_bandwidth(self, version):
        """
        """
        bw  = self.objective_value
        L   = self.extract_ordering(version)
        pos = {u:i for i,u in enumerate(L)}
        dist = lambda a,b:max(pos[a]-pos[b], pos[b]-pos[a])
        return max(dist(u,v) for u,v in self._E)<=bw




    def sumcut(self, version, lowerbound=None, upperbound=None):
        """
        """
        self._init_mip()
        self._all_different(version)
        self._define_lambdas(version)

        y = self.mip.new_variable( name = "y", nonnegative = True)

        
        if version[1:]=='D1' and version[0] in ['A','P','Q','R']:
            for i in range(self._N-1):
                for u in self._V:
                    for v in self.G.neighbor_iterator(u):
                        self.mip.add_constraint( self._lambda_tau(u,v,i) <= y[u,i] )

            self.mip.add_constraint( self.mip.sum( y[u,i] for u in self._V for i in range(self._N-1) )
                                      <= self.opt['opt'] )


        elif version[1:]=='D1' and version[0] in ['S']:
            ll = self.mip.new_variable( name = "l", binary = True, nonnegative = True)
            rr = self.mip.new_variable( name = "l", binary = True, nonnegative = True)
            for i in range(self._N):
                for u in self._V:
                    self.mip.add_constraint( i+1 - self.mip.sum( self.x[w,u] for w in self._V if u!=w )
                                             <= self._N*ll[u,i] )
                    self.mip.add_constraint( self.mip.sum( self.x[w,u] for w in self._V if u!=w ) -i
                                             <= self._N*rr[u,i] )
                    for v in self.G.neighbors(u):
                        self.mip.add_constraint( ll[v,i]+rr[u,i]-1 <= y[u,i] ) 

            self.mip.add_constraint( self.mip.sum( y[u,i] for u in self._V for i in range(self._N-1) )
                                      <= self.opt['opt'] )


        elif version[1:] in ['D2'] and version[0] in ['A','P','Q','R','S']:
            for u,v in self._E:
                self.mip.add_constraint( self._lambda_ell(u,v) <= y[u] )
                self.mip.add_constraint( self._lambda_ell(v,u) <= y[v] )

            self.mip.add_constraint( self.mip.sum( y[u] for u in self._V ) <= self.opt['opt'])


        elif version[1:] in ['D3', 'Pinana04'] and version[0] in ['A','P','Q','R','S']:
            pos = self.mip.new_variable(name="pos", integer=True, nonnegative=True)
            for u in self._V:
                self.mip.add_constraint( self._lambda_pos(u) == pos[u] )
            for u,v in self._E:
                self.mip.add_constraint( pos[u]-pos[v] <= y[u] )
                self.mip.add_constraint( pos[v]-pos[u] <= y[v] )

            self.mip.add_constraint( self.mip.sum( y[u] for u in self._V ) <= self.opt['opt'])

        else:
            raise ValueError("Unknown version: {}.".format(version))


        self._set_bounds(lowerbound=lowerbound, upperbound=upperbound)
        self._set_objective()
        self._solve()

        if self._check and not self._check_sumcut(version):
            raise ValueError("sumcut version {}: something goes wrong...".format(version))
        return self.objective_value, self.extract_ordering(version)

    def _check_sumcut(self, version):
        """
        """
        sc  = self.objective_value
        L   = self.extract_ordering(version)
        pos = {u:i for i,u in enumerate(L)}
        cpt = {u:0 for u in self._V}
        for u,v in self._E:
            if pos[u]>pos[v]:
                cpt[u] = max(cpt[u], pos[u]-pos[v])
            else:
                cpt[v] = max(cpt[v], pos[v]-pos[u])

        return sum(cpt.values())<=sc


    
    def optimal_linear_arrangement(self, version, lowerbound=None, upperbound=None):
        """
        """
        self._init_mip()
        self._all_different(version)
        self._define_lambdas(version)

        y = self.mip.new_variable( name = "y", nonnegative = True, integer = version=='SD1' or version[1:]=='CK' )

        if version[1:] in ['D2'] and version[0] in ['A','P','Q','R','S']:
            for u,v in self._E:
                self.mip.add_constraint( self._lambda_ell(u,v) <= y[u,v] )
                self.mip.add_constraint( self._lambda_ell(v,u) <= y[u,v] )

            self.mip.add_constraint( self.mip.sum( y[u,v] for u,v in self._E ) <= self.opt['opt'] )


        elif version[1:] in ['D3', 'Pinana04'] and version[0] in ['A','P','Q','R','S']:
            pos = self.mip.new_variable(name="pos", integer=True, nonnegative=True)
            for u in self._V:
                self.mip.add_constraint( self._lambda_pos(u) == pos[u] )
            for u,v in self._E:
                self.mip.add_constraint( pos[u]-pos[v] <= y[u,v] )
                self.mip.add_constraint( pos[v]-pos[u] <= y[u,v] )

            self.mip.add_constraint( self.mip.sum( y[u,v] for u,v in self._E ) <= self.opt['opt'] )

            
        elif version[1:] in ['D1'] and version[0] in ['A','P','Q','R']:
            for u,v in self._E:
                for i in range(self._N):
                    self.mip.add_constraint( self._lambda_tau(u,v,i) <= y[u,v,i] )
                    self.mip.add_constraint( self._lambda_tau(v,u,i) <= y[u,v,i] )

            self.mip.add_constraint( self.mip.sum( y[u,v,i] for u,v in self._E for i in range(self._N) )
                                      <= self.opt['opt'] )


        elif version[1:] in ['D1'] and version[0] in ['S']:
            ll = self.mip.new_variable( name = "l", binary = True, nonnegative = True)
            rr = self.mip.new_variable( name = "l", binary = True, nonnegative = True)
            for i in range(self._N):
                for u in self._V:
                    self.mip.add_constraint( i+1 - self.mip.sum( self.x[w,u] for w in self._V if u!=w )
                                                <= self._N*ll[u,i] )
                    self.mip.add_constraint( self.mip.sum( self.x[w,u] for w in self._V if u!=w ) -i
                                                <= self._N*rr[u,i] )
                        
            for i in range(self._N):
                for u,v in self._E:
                    self.mip.add_constraint( ll[u,i]+rr[v,i]-1 <= y[u,v,i] ) 
                    self.mip.add_constraint( ll[v,i]+rr[u,i]-1 <= y[u,v,i] ) 
            
            self.mip.add_constraint( self.mip.sum( y[u,v,i] for u,v in self._E for i in range(self._N) )
                                      <= self.opt['opt'] )


                
        elif version[1:] in ['CK'] and version[0] in ['A','P','Q','R']: # version MILP-20140424
            if version[0]=='A':
                is_at_pos = lambda _v,_i:self.x[_v,_i]
                is_before = lambda _u,_i:self.mip.sum( self.x[_u,_j] for _j in range(_i) )
            else:
                is_at_pos = lambda _v,_i:(self.x[_v,_i]-self.x[_v,_i-1])
                is_before = lambda _u,_i:self.x[_u,_i-1]
            
            diff  = self.mip.new_variable(integer=True, nonnegative=False, name='diff')
            deg   = self.G.degree(labels=True)
            delta = max(deg.values())
            self.mip.add_constraint( self.mip.sum( deg[v]*self.x[v,0] for v in self._V ) == y[0] )
            for i in range(1, self._N):
                for v in self._V:
                    self.mip.add_constraint( deg[v]*is_at_pos(v,i)
                                            - 2*self.mip.sum( is_before(u,i) for u in self.G.neighbors(v) )
                                            - 2*delta*(1-is_at_pos(v,i))
                                            <= diff[i] )

                self.mip.add_constraint( y[i-1] + diff[i] == y[i] )

            self.mip.add_constraint( y[self._N-1] == 0 )

            self.mip.add_constraint( self.mip.sum( y[i] for i in range(self._N) ) <= self.opt['opt'] )


        else:
            raise ValueError("Unknown version: {}.".format(version))
        

        self._set_bounds(lowerbound=lowerbound, upperbound=upperbound)
        self._set_objective()
        self._solve()

        if self._check and not self._check_optimal_linear_arrangement(version):
            raise ValueError("optimal_linear_arrangement version {}: something goes wrong...".format(version))
        return self.objective_value, self.extract_ordering(version)

    def _check_optimal_linear_arrangement(self, version):
        """
        """
        ola  = self.objective_value
        L   = self.extract_ordering(version)
        pos = {u:i for i,u in enumerate(L)}
        dist = lambda a,b:max(pos[a]-pos[b], pos[b]-pos[a])
        return sum(dist(u,v) for u,v in self._E)<=ola

    
    def treewidth(self, version, lowerbound=None, upperbound=None):
        """
        """
        self._init_mip()
        self._all_different(version)

        import itertools

        self.y = self.mip.new_variable( name        = "y",
                                        binary      = version[1:] in ['D1', 'EOF'],
                                        nonnegative = True )

        if version[0] in ['A'] and version[1:] in ['D1']:
            print("WARNING "*10,"\n\n\tnot working yet :((((\n\n","WARNING "*10,"\n\n")
            for u,v in self._E:
                self.mip.add_constraint( self.y[u,v] + self.y[v,u] == 1 )
            
            for u,v,w in itertools.permutations(self._V, 3):
                for i in range(self._N):
                    self.mip.add_constraint( self.y[u,v] + self.y[u,w]
                                             + self.mip.sum( self.x[v,j] for j in range(i+1) )
                                             + self.mip.sum( self.x[w,j] for j in range(i+1,self._N) )
                                             <= self.y[v,w] + 3 )

            for u in self._V:
                self.mip.add_constraint( self.mip.sum( self.y[u,v] for v in self._V if u!=v )
                                         <= self.opt['opt'] )

        elif version[0] in ['P', 'Q'] and version[1:] in ['D1']:
            print("WARNING "*10,"\n\n\tnot working yet :((((\n\n","WARNING "*10,"\n\n")
            for u,v in self._E:
                self.mip.add_constraint( self.y[u,v] + self.y[v,u] == 1 )
            
            for u,v,w in itertools.permutations(self._V,3):
                for i in range(self._N):
                    self.mip.add_constraint( self.y[u,v] + self.y[u,w]
                                             + self.x[v,i]        # v in [0, i]
                                             + (1 - self.x[w,i])  # w in [i+1, n-1]
                                             <= self.y[v,w] + 3 )

            for u in self._V:
                self.mip.add_constraint( self.mip.sum( self.y[u,v] for v in self._V if u!=v )
                                        <= self.opt['opt'] )


        elif version[0] in ['S'] and version[1:] in ['EOF']:
            import itertools

            for u,v in self._E:
                self.mip.add_constraint( self.y[u,v] + self.y[v,u] == 1 )
            
            for u,v in itertools.permutations(self._V, 2):
                self.mip.add_constraint( self.y[u,v] <= self.x[u,v] )

            for u,v in self._E:
                self.mip.add_constraint( self.y[u,v] == self.x[u,v] )
                self.mip.add_constraint( self.y[v,u] == self.x[v,u] )

            for u,v,w in itertools.permutations(self._V, 3):
                self.mip.add_constraint( self.x[v,w] + self.y[u,v] + self.y[u,w] - self.y[v,w] <= 2 )

            for u in self._V:
                self.mip.add_constraint( self.mip.sum( self.y[u,v] for v in self._V ) <= self.opt['opt'] )

        else:
            raise ValueError("Unknown version: {}.".format(version))

        
        self._set_bounds(lowerbound=lowerbound, upperbound=upperbound)
        self._set_objective()
        self._solve()

        #if self._check and not self._check_bandwidth():
        #    raise ValueError("bandwidth version {}: something goes wrong...".format(version))
        return self.objective_value, self.extract_ordering(version)


    
    def minimum_fill_in(self, version, lowerbound=None, upperbound=None):
        """
        """
        self._init_mip()
        self._all_different(version)

        import itertools

        self.y = self.mip.new_variable( name        = "y",
                                        binary      = version[1:] in ['D1', 'EOF'],
                                        nonnegative = True )

        if version[0] in ['A'] and version[1:] in ['D1']:
            print("WARNING "*10,"\n\n\tnot working yet :((((\n\n","WARNING "*10,"\n\n")
            for u,v in self._E:
                self.mip.add_constraint( self.y[u,v] + self.y[v,u] == 1 )

            for u,v,w in itertools.permutations(self._V,3):
                for i in range(self._N):
                    self.mip.add_constraint( self.y[u,v] + self.y[u,w]
                                             + self.mip.sum( self.x[v,j] for j in range(i+1) )
                                             + self.mip.sum( self.x[w,j] for j in range(i+1,self._N) )
                                             <= self.y[v,w] + 3 )

            self.mip.add_constraint( self.mip.sum( self.y[u,v] for u,v in itertools.permutations(self._V,2) )
                                     <= self.opt['opt'] )

        elif version[0] in ['P', 'Q'] and version[1:] in ['D1']:
            print("WARNING "*10,"\n\n\tnot working yet :((((\n\n","WARNING "*10,"\n\n")
            for u,v in self._E:
                self.mip.add_constraint( self.y[u,v] + self.y[v,u] == 1 )

            for u,v,w in itertools.permutations(self._V,3):
                for i in range(self._N):
                    self.mip.add_constraint( self.y[u,v] + self.y[u,w]
                                             + self.x[v,i]        # v in [0, i]
                                             + (1 - self.x[w,i])  # w in [i+1, n-1]
                                             <= self.y[v,w] + 3 )

            self.mip.add_constraint( self.mip.sum( self.y[u,v] for u,v in itertools.permutations(self._V,2) )
                                     <= self.opt['opt'] )


        elif version[0] in ['S'] and version[1:] in ['EOF']:
            import itertools

            for u,v in self._E:
                self.mip.add_constraint( self.y[u,v] + self.y[v,u] == 1 )

            for u,v in itertools.permutations(self._V, 2):
                self.mip.add_constraint( self.y[u,v] <= self.x[u,v] )

            for u,v in self._E:
                self.mip.add_constraint( self.y[u,v] == self.x[u,v] )
                self.mip.add_constraint( self.y[v,u] == self.x[v,u] )

            for u,v,w in itertools.permutations(self._V, 3):
                self.mip.add_constraint( self.x[v,w] + self.y[u,v] + self.y[u,w] - self.y[v,w] <= 2 )

            self.mip.add_constraint( self.mip.sum( self.y[u,v] for u,v in itertools.permutations(self._V, 2 ) )
                                     <= self.opt['opt'] )

        else:
            raise ValueError("Unknown version: {}.".format(version))

        
        self._set_bounds(lowerbound=lowerbound, upperbound=upperbound)
        self._set_objective()
        self._solve()

        #if self._check and not self._check_bandwidth():
        #    raise ValueError("bandwidth version {}: something goes wrong...".format(version))
        return self.objective_value, self.extract_ordering(version)




    def pathwidth_BBNNPR(self):
        """
        """
        self._init_mip()

        R = range(1,self._N+1)
        Rbar = range(0,self._N+2)

        x  = self.mip.new_variable(name="x", binary=True, nonnegative=True)   #1
        b  = self.mip.new_variable(name="x", binary=True, nonnegative=True)   #2
        e  = self.mip.new_variable(name="x", binary=True, nonnegative=True)   #3
        xe = self.mip.new_variable(name="x", binary=True, nonnegative=True)   #10

        for v in self._V:  #4
            self.mip.add_constraint( x[v,0] == 0 )
            self.mip.add_constraint( x[v,self._N+1] == 0 )

        for v in self._V: #5
            self.mip.add_constraint( self.mip.sum( x[v,i] for i in Rbar ) >= 1 )

        for v in self._V: #6  #7
            self.mip.add_constraint( self.mip.sum( b[v,i] for i in R ) == 1 )
            self.mip.add_constraint( self.mip.sum( e[v,i] for i in R ) == 1 )

        for v in self._V: #8 #9
            for i in R:
                self.mip.add_constraint( x[v,i-1]+b[v,i] >= x[v,i] )
                self.mip.add_constraint( x[v,i] <= x[v,i+1]+e[v,i] )

        for u,v in self._E: #11
            self.mip.add_constraint( self.mip.sum( xe[u,v,i] for i in R ) >= 1 )

        for u,v in self._E: #12 #13
            for i in R:
                self.mip.add_constraint( xe[u,v,i] <= x[u,i] )
                self.mip.add_constraint( xe[u,v,i] <= x[v,i] )

        for i in R:
            self.mip.add_constraint( self.mip.sum( x[v,i] for v in self._V ) <= self.opt['opt']+1 )

        self._set_objective()
        self._solve()

        if False:
            valx = self.mip.get_values(x)
            for v in self._V:
                st = "{}:\t".format(v)
                for i in R:
                    st += "1 " if valx[v,i]>0 else "0 "
                print(st)
            
        return self.objective_value # , self.extract_ordering(version)

    def bandwidth_BBNNPR(self, show_log=False):
        """
        """
        self._init_mip()
        self._version = 'bwBBNNPR'

        R = list(range(self._N))
        Rbar = [R[0]-1] + R + [R[-1]+1]

        x  = self.mip.new_variable(name="x", binary=True, nonnegative=True)   #1
        b  = self.mip.new_variable(name="x", binary=True, nonnegative=True)   #2
        e  = self.mip.new_variable(name="x", binary=True, nonnegative=True)   #3

        for v in self._V:  #4
            self.mip.add_constraint( x[v,Rbar[0]] == 0 )
            self.mip.add_constraint( x[v,Rbar[-1]] == 0 )

        for v in self._V: #5
            self.mip.add_constraint( self.mip.sum( x[v,i] for i in Rbar ) >= 1 )

        for v in self._V: #6  #7
            self.mip.add_constraint( self.mip.sum( b[v,i] for i in R ) == 1 )
            self.mip.add_constraint( self.mip.sum( e[v,i] for i in R ) == 1 )

        for v in self._V: #8 #9
            for i in R:
                self.mip.add_constraint( x[v,i-1]+b[v,i] >= x[v,i] )
                self.mip.add_constraint( x[v,i] <= x[v,i+1]+e[v,i] )

        for i in R: #14
            self.mip.add_constraint( self.mip.sum( x[v,i] for v in self._V ) <= 1 )

        nb_const = self.mip.number_of_constraints()

        LB = ceil(max(self.G.degree())/2)
        UB = self._N-1
        first = True
        t = walltime()
        while LB<UB:
            if UB==LB+1:
                if first:
                    k = LB
                    first = False
                else:
                    k = LB = UB
                    break
            else:
                k = floor((LB+UB)/2)

            if self._verbose:
                print("try with k = {}  {} {}".format(k,LB,UB))

            for u,v in self._E: #15
                for i in R:
                    self.mip.add_constraint( x[u,i] <= self.mip.sum( x[v,j] for j in range(i-k, i+k+1) if j in Rbar ) )

            try:
                self.mip.solve(log=show_log)
                UB = k
            except:
                self.mip.remove_constraints(range(nb_const, self.mip.number_of_constraints()))
                LB = k

            if walltime()-t>=self._timelimit:
                break

        self.time = walltime()-t
        self.objective_value = UB
        self._best_objective_value = LB
        self._relative_mip_gap = float((UB-LB)/UB)

        if False:
            valx = self.mip.get_values(x)
            for v in self._V:
                st = "{}:\t".format(v)
                for i in R:
                    st += "1 " if valx[v,i]>0 else ". "
                print(st)
        
        #if self._check and not self._check_bandwidth('AA'):
        #    raise ValueError("bandwidth version {}: something goes wrong...".format('BBNNPR'))
        return self.objective_value #, self.extract_ordering('AA')
   




################################################################################
# 
################################################################################
def graph_to_dzn(G, filename, verbose=True):
    st = "% {}\nn = {};\nm = {};\nvertices = 0..n-1;\nedges = [|\n".format(G.name(),G.order(),G.size())
    map = {u:i for i,u in enumerate(G.vertices())}
    ord = lambda x,y: (x,y) if x<y else (y,x)
    E = sorted( ord(map[u], map[v]) for u,v in G.edges(labels=None) )
    for u,v in E[:-1]:
        st += "{}, {}|\n".format(u,v)
    st += "{}, {}|];\n".format(E[-1][0], E[-1][1])
    if verbose:
        print(st)
    with open(filename, 'w') as f:
        f.write(st)
    
    
